"""
Record shell history.

This is a command to be called from shell-specific hooks.
This Python implementation is a reference implementation.
Probably it makes sense to write this command naively in
shells to make it faster.

The dumped data goes under the ``~/.config/rash/data/record``
directory.

"""

# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import time
import json

from .utils.pathutils import mkdirp
from .utils.py3compat import getcwd
from .config import ConfigStore


def get_tty():
    """
    Return \"os.ttyname(0 or 1 or 2)\".
    """
    for i in range(3):
        try:
            return os.ttyname(i)
            break
        except OSError:
            pass


def get_environ(keys):
    """
    Get environment variables from :data:`os.environ`.

    :type keys: [str]
    :rtype: dict

    Some additional features.

    * If 'HOST' is not in :data:`os.environ`, this function
      automatically fetch it using :meth:`platform.node`.
    * If 'TTY' is not in :data:`os.environ`, this function
      automatically fetch it using :meth:`os.ttyname`.
    * Set 'RASH_SPENV_TERMINAL' if needed.

    """
    items = ((k, os.environ.get(k)) for k in keys)
    subenv = dict((k, v) for (k, v) in items if v is not None)
    needset = lambda k: k in keys and not subenv.get(k)

    def setifnonempty(key, value):
        if value:
            subenv[key] = value

    if needset('HOST'):
        import platform
        subenv['HOST'] = platform.node()
    if needset('TTY'):
        setifnonempty('TTY', get_tty())
    if needset('RASH_SPENV_TERMINAL'):
        from .utils.termdetection import detect_terminal
        setifnonempty('RASH_SPENV_TERMINAL', detect_terminal())
    return subenv


def generate_session_id(data):
    """
    Generate session ID based on HOST, TTY, PID [#]_ and start time.

    :type data: dict
    :rtype: str

    .. [#] PID of the shell, i.e., PPID of this Python process.

    """
    host = data['environ']['HOST']
    tty = data['environ'].get('TTY') or 'NO_TTY'
    return ':'.join(map(str, [
        host, tty, os.getppid(), data['start']]))


def record_run(record_type, print_session_id, **kwds):
    """
    Record shell history.
    """
    if print_session_id and record_type != 'init':
        raise RuntimeError(
            '--print-session-id should be used with --record-type=init')

    # SOMEDAY: make environment variables to log configurable
    if record_type == 'init':
        envkeys = [
            'SHELL', 'TERM', 'HOST', 'TTY', 'USER', 'DISPLAY',
            # SOMEDAY: Reevaluate if "RASH_SPENV_TERMINAL" is the right choice.
            # Here, I am using `environ` dict as a generic key value store.
            # Using 'RASH_SPENV_' as a prefix key, it is very easy to add
            # new variable to track.
            'RASH_SPENV_TERMINAL',
        ]
    elif record_type == 'exit':
        envkeys = []
    elif record_type == 'command':
        envkeys = ['PATH']

    conf = ConfigStore()
    json_path = os.path.join(conf.record_path,
                             record_type,
                             time.strftime('%Y-%m-%d-%H%M%S.json'))
    mkdirp(os.path.dirname(json_path))

    # Command line options directly map to record keys
    data = dict((k, v) for (k, v) in kwds.items() if v is not None)
    data.update(
        environ=get_environ(envkeys),
    )

    # Automatically set some missing variables:
    data.setdefault('cwd', getcwd())
    if record_type in ['command', 'exit']:
        data.setdefault('stop', int(time.time()))
    elif record_type in ['init']:
        data.setdefault('start', int(time.time()))

    if print_session_id:
        data['session_id'] = generate_session_id(data)
        print(data['session_id'])

    with open(json_path, 'w') as fp:
        json.dump(data, fp)


def record_add_arguments(parser):
    parser.add_argument(
        '--record-type', default='command',
        choices=['command', 'init', 'exit'],
        help='type of record to store.')
    parser.add_argument(
        '--command',
        help="command that was ran.")
    parser.add_argument(
        '--cwd',
        help='''
        Like $PWD, but callee can set it to consider command that
        changes directory (e.g., cd).
        ''')
    parser.add_argument(
        '--exit-code', type=int,
        help="exit code $? of the command.")
    parser.add_argument(
        '--pipestatus', type=int, nargs='+',
        help="$pipestatus (zsh) / $PIPESTATUS (bash)")
    parser.add_argument(
        '--start', type=int,
        help='the time COMMAND is started.')
    parser.add_argument(
        '--stop', type=int,
        help='the time COMMAND is finished.')
    parser.add_argument(
        '--session-id',
        help='''
        RASH session ID generated by --print-session-id.
        This option should be used with `command` or `exit` RECORD_TYPE.
        ''')
    parser.add_argument(
        '--print-session-id', default=False, action='store_true',
        help='''
        print generated session ID to stdout.
        This option should be used with `init` RECORD_TYPE.
        ''')


commands = [
    ('record', record_add_arguments, record_run),
]



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import json
import warnings

from .database import DataBase


class Indexer(object):

    """
    Translate JSON files into SQLite DB.
    """

    def __init__(self, conf, check_duplicate, keep_json, record_path=None):
        """
        Create an indexer.

        :type            conf: rash.config.ConfigStore
        :arg             conf:
        :type check_duplicate: bool
        :arg  check_duplicate: See :meth:`DataBase.import_dict`.
        :type       keep_json: bool
        :arg        keep_json: Do not remove JSON files.
                               Imply ``check_duplicate=True``.
        :type     record_path: str or None
        :arg      record_path: Default to `conf.record_path`.

        """
        from .log import logger
        self.logger = logger
        if keep_json:
            check_duplicate = True
        self.conf = conf
        self.check_duplicate = check_duplicate
        self.keep_json = keep_json
        self.record_path = record_path or conf.record_path
        self.db = DataBase(conf.db_path)
        if record_path:
            self.check_path(record_path, '`record_path`')

        self.logger.debug('Indexer initialized')
        self.logger.debug('check_duplicate = %r', self.check_duplicate)
        self.logger.debug('keep_json = %r', self.keep_json)
        self.logger.debug('record_path = %r', self.record_path)

    def get_record_type(self, path):
        relpath = os.path.relpath(path, self.conf.record_path)
        dirs = relpath.split(os.path.sep, 1)
        return dirs[0] if dirs else None

    def check_path(self, path, name='path'):
        if self.get_record_type(path) not in ['command', 'init', 'exit']:
            raise RuntimeError(
                '{0} must be under {1}'.format(
                    name,
                    os.path.join(self.conf.record_path,
                                 '{command,init,exit}',
                                 '')))

    def index_record(self, json_path):
        """
        Import `json_path` and remove it if :attr:`keep_json` is false.
        """
        self.logger.debug('Indexing record: %s', json_path)
        json_path = os.path.abspath(json_path)
        self.check_path(json_path, '`json_path`')

        with open(json_path) as fp:
            try:
                dct = json.load(fp)
            except ValueError:
                warnings.warn(
                    'Ignoring invalid JSON file at: {0}'.format(json_path))
                return

        record_type = self.get_record_type(json_path)
        kwds = {}
        if record_type == 'command':
            importer = self.db.import_dict
            kwds.update(check_duplicate=self.check_duplicate)
        elif record_type == 'init':
            importer = self.db.import_init_dict
        elif record_type == 'exit':
            importer = self.db.import_exit_dict
        else:
            raise ValueError("Unknown record type: {0}".format(record_type))
        importer(dct, **kwds)

        if not self.keep_json:
            self.logger.info('Removing JSON record: %s', json_path)
            os.remove(json_path)

    def find_record_files(self):
        """
        Yield paths to record files.
        """
        for (root, _, files) in os.walk(self.record_path):
            for f in (f for f in files if f.endswith('.json')):
                yield os.path.join(root, f)

    def index_all(self):
        """
        Index all records under :attr:`record_path`.
        """
        self.logger.debug('Start indexing all records under: %s',
                          self.record_path)
        with self.db.connection():
            for json_path in sorted(self.find_record_files()):
                self.index_record(json_path)



# [[[cog import cog; cog.outl('"""\n%s\n"""' % file('../README.rst').read())]]]
"""
=============================
 Rash Advances Shell History
=============================


.. sidebar:: Links:

   * `Repository <https://github.com/tkf/rash>`_ (at GitHub)
   * `Issue tracker <https://github.com/tkf/rash/issues>`_ (at GitHub)
   * `PyPI <http://pypi.python.org/pypi/rash>`_
   * `Travis CI <https://travis-ci.org/#!/tkf/rash>`_ |build-status|


What is this?
=============

Shell history is useful.  But it can be more useful if it logs more
data points.  For example, if you forget which `make` target to run
for certain project, you'd want to search shell commands that are
ran in particular directory.  Wouldn't it be nice if you can do this?::

   rash search --cwd . "make*"

RASH records many data points and they are stored in SQLite database.
Here is a list of recorded information [#]_.

#. Current directory (``$PWD``).
#. Exit code (``$?``)
#. Exit code of pipes (``$PIPESTATUS`` / ``$pipestatus``)
#. The time command is started and terminated.
#. Environment variable (``$PATH``, ``$SHELL``, ``$TERM``, ``$HOST``, etc.)
#. Real terminal.  ``$TERM`` is used to fake programs.
   RASH can detect if you are in tmux, byobu, screen, gnome-terminal, etc.
#. Session information.  If you go back and forth in some terminals,
   RASH does not loose in which sequence you ran the commands in which
   terminal.

.. [#] If you are curious, checkout ``rash record --help``.


Install
=======

RASH is written in Python.  The easiest way to install is to use `pip`
(or `easy_install`, if you wish).  You may need `sudo` for installing
it in a system directory.::

   pip install rash

RASH tested against Python 2.6, 2.7 and 3.2.  However, as watchdog_
does not work with Python 3, you can't get full power of RASH with
Python 3.

If you want to use developmental version, just clone the git repository
and add the following in your RC file.::

  alias rash="PATH/TO/RASH/rash_cli.py"


Setup
=====
Add this to your `.zshrc` or `.bashrc`.  That's all.::

   eval "$(rash init)"

For more information, see ``rash init --help``.


Usage
=====

Searching history -- ``rash search``
------------------------------------

After your shell history is accumulated by RASH, it's the time to
make use of the history!  See ``rash search --help`` for detailed
information.  Here is some examples.

Forget how to run automated test for the current project?::

   rash search --cwd . "*test*" "tox*"

All git commands you ran in one week.::

   rash search --time-after "1 week ago" "git*"

Some intensive task you ran in the current project that succeeded and
took longer than 30 minutes.::

   rash search --cwd-under . --include-exit-code 0 --duration-longer-than 30m

**NOT IMPLEMENTED**
What did I do after `cd`-ing to some directory?::

   rash search --after-context 5 "cd SOME-DIRECTORY"

All failed commands you ran at this directory.::

   rash search --cwd . --exclude-exit-code 0

**NOT IMPLEMENTED**
Top 5 programs you use most.::

   rash search --limit 5 --sort-by-program-frequency

Count number of commands you ran in one day::

   rash search --limit -1 --no-unique --time-after "1 day ago" | wc -l


Showing detailed information -- ``rash show``
---------------------------------------------

If you give ``--with-command-id`` to ``rash search`` command, it prints out
ID number for each command history.::

   % rash search --with-command-id --limit 5 "*git*"
    359  git log
   1253  git help clone
   1677  git help diff
   1678  git diff --word-diff
   1780  git merge

You can see all information associated with a command with
``rash show`` command::

   rash show 1677


Interactive search -- ``rash isearch`` (experimental!)
------------------------------------------------------

Searching history using command line is not fast.
You can use ``rash isearch`` command to interactively search
history and see the result immediately as you type.

You need percol_ to use this feature.

Zsh user can setup a keybind like this::

  # Type `Ctrl-x r` to start isearch
  bindkey "^Xr" rash-zle-isearch

Defining this function in your rc file can be handy and it is
usable for bash users.::

  rash-isearch(){
    eval "$(rash isearch)"
  }

.. _percol: https://github.com/mooz/percol


Dependency
==========

Python modules:

* watchdog_ [#nopy3k]_
* parsedatetime_ [#nopy3k]_

.. _watchdog: http://pypi.python.org/pypi/watchdog/
.. _parsedatetime: http://pypi.python.org/pypi/parsedatetime/

.. [#nopy3k] These modules do not support Python 3.
             They are not installed in if you use Python 3
             and related functionality is disabled.

Platforms
---------

UNIX-like systems
  RASH is tested in Linux and I am using in Linux.
  It should work in other UNIX-like systems like BSD.

Mac OS
  I guess it works.  Not tested.

MS Windows
  Probably no one wants to use a shell tool in windows, but I
  try to avoid stuff that is platform specific.  Only the
  daemon launcher will not work on Windows but there is several
  ways to avoid using it.  See ``rash init --help``.


Design principle
================

RASH's design is focused on sparseness.  There are several stages
of data transformation until you see the search result, and they
are done by separated processes.

First, `rash record` command dumps shell history in raw JSON record.
This part of program does not touches to DB to make process very fast.
As there is no complex transformation in this command, probably in the
future version is is better to rewrite it entirely in shell function.

Second, `rash daemon` runs in background and watches the directory to
store JSON record.  When JSON record arrives, it insert the data into
database.

`rash record` and `rash daemon` are setup by simple shell snippet
``eval $(rash init)``.

Finally, you can search through command history using search interface
such as `rash search`.  This search is very fast as you don't read
all JSON records in separated files.

::

   +-------+         +--------+         +--------+         +--------+
   | Shell |         | Raw    |         | SQLite |         | Search |
   | hooks |-------->| JSON   |-------->|   DB   |-------->| result |
   +-------+         | record |         +--------+         +--------+
                     +--------+

           `rash record`      `rash daemon`      `rash search`
                                                  `rash show`

           \------------------------------/      \------------/
              `rash init` setups them           search interface

License
=======

RASH is licensed under GPL v3.
See COPYING for details.


.. Travis CI build status badge
.. |build-status|
   image:: https://secure.travis-ci.org/tkf/rash.png?branch=master
   :target: http://travis-ci.org/tkf/rash
   :alt: Build Status

"""
# [[[end]]]

# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


__version__ = '0.1.0'
__author__ = 'Takafumi Arakaki'
__license__ = 'GNU General Public License v3 (GPLv3)'



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import tempfile
import shutil
import json

from ..config import ConfigStore
from ..indexer import Indexer
from ..utils.pathutils import mkdirp
from .utils import BaseTestCase


class TestIndexer(BaseTestCase):

    def setUp(self):
        self.base_path = tempfile.mkdtemp(prefix='rash-test-')
        self.conf = ConfigStore(self.base_path)

    def tearDown(self):
        shutil.rmtree(self.base_path)

    def get_indexer(self, keep_json=True, check_duplicate=True):
        return Indexer(self.conf, check_duplicate, keep_json)

    def prepare_records(self, **records):
        if set(records) > set(['command', 'init', 'exit']):
            raise ValueError(
                'Unknown record type in {0}'.format(list(records)))
        paths = []
        for (rectype, data_list) in records.items():
            for (i, data) in enumerate(data_list):
                json_path = os.path.join(self.conf.record_path,
                                         rectype,
                                         '{0:05d}.json'.format(i))
                mkdirp(os.path.dirname(json_path))
                with open(json_path, 'w') as f:
                    json.dump(data, f)
                paths.append(json_path)
        return paths

    def get_dummy_records(self, num_command=1, num_init=1, num_exit=1):
        gen = lambda i, **kwds: dict(session_id='SID-{0}'.format(i), **kwds)
        return dict(
            command=[gen(i, start=i, stop=i + 1) for i in range(num_command)],
            init=[gen(i, start=i) for i in range(num_init)],
            exit=[gen(i, stop=i) for i in range(num_exit)],
        )

    def test_find_record_files(self):
        indexer = self.get_indexer()
        self.assertEqual(list(indexer.find_record_files()), [])
        desired_paths = self.prepare_records(**self.get_dummy_records())
        actual_paths = list(indexer.find_record_files())
        self.assertSetEqual(set(actual_paths), set(desired_paths))

    def test_index_all_and_keep_json(self):
        desired_paths = self.prepare_records(**self.get_dummy_records())
        indexer = self.get_indexer()
        indexer.index_all()
        actual_paths = list(indexer.find_record_files())
        self.assertSetEqual(set(actual_paths), set(desired_paths))

    def test_index_all_and_discard_json(self):
        self.prepare_records(**self.get_dummy_records())
        indexer = self.get_indexer(keep_json=False)
        indexer.index_all()
        actual_paths = list(indexer.find_record_files())
        self.assertEqual(actual_paths, [])



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os


def shell_name(shell):
    return shell.rsplit(os.path.sep, 1)[-1]


def find_init(shell):
    rash_dir = os.path.dirname(__file__)
    return os.path.join(rash_dir, 'ext', 'rash.{0}'.format(shell_name(shell)))


def init_run(shell, no_daemon, daemon_options, daemon_outfile):
    """
    Configure your shell.

    Add the following line in your shell RC file and then you are
    ready to go::

      eval $(%(prog)s)

    To check if your shell is supported, simply run::

      %(prog)s --no-daemon

    By default, this command also starts daemon in background to
    automatically inndex shell history records.  To not start daemon,
    use --no-daemon option like this::

      eval $(%(prog)s --no-daemon)

    To see the other methods to launch the daemon process, see
    ``rash daemon --help``.

    """
    init_file = find_init(shell)
    if os.path.exists(init_file):
        print("source '{0}'".format(init_file))
    else:
        raise RuntimeError(
            "Shell '{0}' is not supported.".format(shell_name(shell)))

    if not no_daemon:
        from .daemon import start_daemon_in_subprocess
        start_daemon_in_subprocess(daemon_options, daemon_outfile)


def init_add_arguments(parser):
    parser.add_argument(
        '--shell', default=os.environ.get('SHELL'),
        help="""
        name of shell you are using.  directory before the last /
        is discarded.  It defaults to $SHELL.
        """)
    parser.add_argument(
        '--no-daemon', action='store_true', default=False,
        help="""
        Do not start daemon.  By default, daemon is started if
        there is no already running daemon.
        """)
    parser.add_argument(
        '--daemon-opt', dest='daemon_options', action='append', default=[],
        help="""
        Add options given to daemon.  See "rash daemon --help" for
        available options.  It can be specified many times.
        Note that --no-error is always passed to the daemon command.
        """)
    parser.add_argument(
        '--daemon-outfile', default=os.devnull,
        help="""
        Path to redirect STDOUT and STDERR of daemon process.
        This is mostly for debugging.
        """)


commands = [
    ('init', init_add_arguments, init_run),
]



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
import signal

try:
    from watchdog.observers import Observer
    from watchdog.events import (
        FileSystemEventHandler, FileCreatedEvent)
    assert FileSystemEventHandler  # fool pyflakes
except ImportError:
    # Dummy class for making this module importable:
    FileSystemEventHandler = object


class RecordHandler(FileSystemEventHandler):

    def __init__(self, indexer, **kwds):
        self.__indexer = indexer
        super(RecordHandler, self).__init__(**kwds)

    def on_created(self, event):
        if isinstance(event, FileCreatedEvent):
            self.__indexer.index_record(event.src_path)


def raise_keyboardinterrupt(_signum, _frame):
    raise KeyboardInterrupt


def install_sigterm_handler():
    signal.signal(signal.SIGTERM, raise_keyboardinterrupt)


def watch_record(indexer):
    """
    Start watching `conf.record_path`.

    :type indexer: rash.indexer.Indexer

    """

    event_handler = RecordHandler(indexer)
    observer = Observer()
    observer.schedule(event_handler, path=indexer.record_path, recursive=True)
    indexer.logger.debug('Start observer.')
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        indexer.logger.debug('Got KeyboardInterrupt. Stopping observer.')
        observer.stop()
    indexer.logger.debug('Joining observer.')
    observer.join()
    indexer.logger.debug('Finish watching record.')



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging


def loglevel(level):
    """
    Convert any representation of `level` to an int appropriately.

    :type level: int or str
    :rtype: int

    >>> loglevel('DEBUG') == logging.DEBUG
    True
    >>> loglevel(10)
    10
    >>> loglevel(None)
    Traceback (most recent call last):
      ...
    ValueError: None is not a proper log level.

    """
    if isinstance(level, str):
        level = getattr(logging, level.upper())
    elif isinstance(level, int):
        pass
    else:
        raise ValueError('{0!r} is not a proper log level.'.format(level))
    return level


logger = logging.getLogger('rash')


def setup_daemon_log_file(conf):
    """
    Attach file handler to RASH logger.

    :type conf: rash.config.ConfigStore

    """
    level = loglevel(conf.daemon_log_level)
    handler = logging.FileHandler(filename=conf.daemon_log_path)
    handler.setLevel(level)
    logger.setLevel(level)
    logger.addHandler(handler)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest
import functools
from contextlib import contextmanager


class BaseTestCase(unittest.TestCase):

    if not hasattr(unittest.TestCase, 'assertIn'):
        def assertIn(self, member, container):
            self.assertTrue(member in container)

    if not hasattr(unittest.TestCase, 'assertNotIn'):
        def assertNotIn(self, member, container):
            self.assertTrue(member not in container)

    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, obj, cls, msg=None):
            self.assertTrue(isinstance(obj, cls), msg)

    if not hasattr(unittest.TestCase, 'assertSetEqual'):
        def assertSetEqual(self, set1, set2, msg=None):
            self.assertEqual(set1, set2, msg)

try:
    skipIf = unittest.skipIf
except AttributeError:

    def skipIf(condition, reason):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwds):
                if condition:
                    print("Skipping {0} because:".format(func.__name__))
                    print(reason)
                else:
                    return func(*args, **kwds)
            return wrapper
        return decorator


@contextmanager
def monkeypatch(obj, name, attr):
    """
    Context manager to replace attribute named `name` in `obj` with `attr`.
    """
    orig = getattr(obj, name)
    setattr(obj, name, attr)
    yield
    setattr(obj, name, orig)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def isearch_run(**kwds):
    r"""
    Interactive history search that updated as you type.

    The query for this program is the same as the one for
    ``rash search`` command.

    You need percol_ to use this command.

    _percol: https://github.com/mooz/percol

    If you use zsh, you can setup a keybind like this to quickly
    launch iserch and execute the result.::

      # Type `Ctrl-x r` to start isearch
      bindkey "^Xr" rash-zle-isearch

    If you like command or you are not using zsh, you can add
    something like the following in your rc file to start and
    execute the chosen command.

      rash-isearch(){
        eval "$(rash isearch)"
      }

    To pass long and complex query, give them after "--",
    like this.::

      rash isearch -- \
        --cwd . \
        --exclude-pattern "*rash *" \
        --include-pattern "*test*" \
        --include-pattern "tox*" \
        --include-pattern "make *test*"

    """
    from .config import ConfigStore
    from .interactive_search import launch_isearch
    launch_isearch(ConfigStore(), **kwds)


def isearch_add_arguments(parser):
    parser.add_argument(
        '--query', '-q', default='',
        help='default query')
    parser.add_argument(
        'base_query', nargs='*', default=[],
        help="""
        The part of query that is not shown in UI and is impossible
        to rewrite in this session.  Useful for putting long and
        complex query.
        """)


commands = [
    ('isearch', isearch_add_arguments, isearch_run),
]



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os


def subdict_by_key_prefix(dct, prefix):
    items = []
    for (k, v) in dct.items():
        if k.startswith(prefix):
            items.append((k, v))
    return dict(items)


def detect_terminal(_environ=os.environ):
    """
    Detect "terminal" you are using.

    First, this function checks if you are in tmux, byobu, or screen.
    If not it uses $COLORTERM [#]_ if defined and fallbacks to $TERM.

    .. [#] So, if you are in Gnome Terminal you have "gnome-terminal"
       instead of "xterm-color"".

    """
    if _environ.get('TMUX'):
        return 'tmux'
    elif subdict_by_key_prefix(_environ, 'BYOBU'):
        return 'byobu'
    elif _environ.get('TERM').startswith('screen'):
        return _environ['TERM']
    elif _environ.get('COLORTERM'):
        return _environ['COLORTERM']
    else:
        return _environ.get('TERM')


if __name__ == '__main__':
    print(detect_terminal())



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def show_run(command_history_id):
    """
    Show detailed command history by its ID.
    """
    from pprint import pprint
    from .config import ConfigStore
    from .database import DataBase
    db = DataBase(ConfigStore().db_path)
    with db.connection():
        for ch_id in command_history_id:
            crec = db.get_full_command_record(ch_id)
            pprint(crec.__dict__)
            print("")


def show_add_arguments(parser):
    parser.add_argument(
        'command_history_id', nargs='+', type=int,
        help="""
        Integer ID of command history.
        """)


commands = [
    ('show', show_add_arguments, show_run),
]



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def index_run(record_path, keep_json, check_duplicate):
    """
    Convert raw JSON records into sqlite3 DB.

    Normally RASH launches a daemon that takes care of indexing.
    See ``rash daemon --help``.

    """
    from .config import ConfigStore
    from .indexer import Indexer
    conf = ConfigStore()
    indexer = Indexer(conf, check_duplicate, keep_json, record_path)
    indexer.index_all()


def index_add_arguments(parser):
    parser.add_argument(
        'record_path', nargs='?',
        help="""
        specify the directory that has JSON records.
        """)
    parser.add_argument(
        '--keep-json', default=False, action='store_true',
        help="""
        Do not remove old JSON files.  It turns on --check-duplicate.
        """)
    parser.add_argument(
        '--check-duplicate', default=False, action='store_true',
        help='do not store already existing history in DB.')


commands = [
    ('index', index_add_arguments, index_run),
]



"""
Command line interface.
"""

import argparse
import textwrap

EPILOG = """
# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


class Formatter(argparse.RawDescriptionHelpFormatter,
                argparse.ArgumentDefaultsHelpFormatter):
    pass


def get_parser(commands):
    """
    Generate argument parser given a list of subcommand specifications.

    :type commands: list of (str, function, function)
    :arg  commands:
        Each element must be a tuple ``(name, adder, runner)``.

        :param   name: subcommand
        :param  adder: a function takes one object which is an instance
                       of :class:`argparse.ArgumentParser` and add
                       arguments to it
        :param runner: a function takes keyword arguments which must be
                       specified by the arguments parsed by the parser
                       defined by `adder`.  Docstring of this function
                       will be the description of the subcommand.

    """
    parser = argparse.ArgumentParser(
        formatter_class=Formatter,
        description=__doc__,
        epilog=EPILOG,
    )
    subparsers = parser.add_subparsers()

    for (name, adder, runner) in commands:
        subp = subparsers.add_parser(
            name,
            formatter_class=Formatter,
            description=runner.__doc__ and textwrap.dedent(runner.__doc__))
        adder(subp)
        subp.set_defaults(func=runner)

    return parser


def main(args=None):
    from . import init
    from . import record
    from . import daemon
    from . import search
    from . import show
    from . import index
    from . import isearch
    # from . import MODULE
    parser = get_parser(
        init.commands
        + record.commands
        + daemon.commands
        + search.commands
        + show.commands
        + index.commands
        + isearch.commands
        # + MODULE.commands
    )
    ns = parser.parse_args(args=args)
    applyargs = lambda func, **kwds: func(**kwds)
    applyargs(**vars(ns))


if __name__ == '__main__':
    main()



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
PY3 = (sys.version_info[0] >= 3)

try:
    getcwd = os.getcwd
except AttributeError:
    getcwd = os.getcwdu


try:
    from contextlib import nested
except ImportError:
    from contextlib import contextmanager

    @contextmanager
    def nested(*managers):
        if managers:
            with managers[0] as ctx:
                with nested(*managers[1:]) as rest:
                    yield (ctx,) + rest
        else:
            yield ()



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sqlite3
from contextlib import closing, contextmanager
import datetime
import warnings
import itertools

from .utils.iterutils import nonempty, repeat
from .model import CommandRecord, SessionRecord, VersionRecord

schema_version = '0.1'


def concat_expr(operator, conditions):
    """
    Concatenate `conditions` with `operator` and wrap it by ().

    It returns a string in a list or empty list, if `conditions` is empty.

    """
    expr = " {0} ".format(operator).join(conditions)
    return ["({0})".format(expr)] if expr else []


def convert_ts(ts):
    """
    Convert timestamp (ts)

    :type ts: int or str or None
    :arg  ts: Unix timestamp
    :rtype: datetime.datetime or str or None

    """
    if ts is None:
        return None
    try:
        return datetime.datetime.utcfromtimestamp(ts)
    except TypeError:
        pass
    return ts


def normalize_directory(path):
    """
    Append "/" to `path` if needed.
    """
    if path is None:
        return None
    if path.endswith(os.path.sep):
        return path
    else:
        return path + os.path.sep


class DataBase(object):

    schemapath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'schema.sql')

    def __init__(self, dbpath):
        self.dbpath = dbpath
        if not os.path.exists(dbpath):
            self._init_db()
        self.update_version_records()

    def _get_db(self):
        """Returns a new connection to the database."""
        return closing(sqlite3.connect(self.dbpath))

    def _init_db(self):
        """Creates the database tables."""
        with self._get_db() as db:
            with open(self.schemapath) as f:
                db.cursor().executescript(f.read())
            db.commit()

    @contextmanager
    def connection(self, commit=False):
        """
        Context manager to keep around DB connection.

        :rtype: sqlite3.Connection

        SOMEDAY: Get rid of this function.  Keeping connection around as
        an argument to the method using this context manager is
        probably better as it is more explicit.

        """
        if commit:
            self._need_commit = True
        if self._db:
            yield self._db
        else:
            try:
                with self._get_db() as db:
                    self._db = db
                    yield self._db
                    if self._need_commit:
                        db.commit()
            finally:
                self._db = None
                self._need_commit = False
    _db = None
    _need_commit = False

    def close_connection(self):
        """
        Close connection kept by :meth:`connection`.

        If commit is needed, :meth:`sqlite3.Connection.commit`
        is called first and then :meth:`sqlite3.Connection.interrupt`
        is called.

        A few methods/generators support :meth:`close_connection`:

        - :meth:`search_command_record`
        - :meth:`select_by_command_record`

        """
        if self._db:
            db = self._db
            try:
                if self._need_commit:
                    db.commit()
            finally:
                db.interrupt()
                self._db = None
                self._need_commit = False

    def _executing(self, sql, params=[]):
        """
        Execute and yield rows in a way to support :meth:`close_connection`.
        """
        with self.connection() as connection:
            for row in connection.execute(sql, params):
                yield row
                if not self._db:
                    return

    def _select_rows(self, rowclass, keys, sql, params):
        return (rowclass(**dict(zip(keys, row)))
                for row in self._executing(sql, params))

    def get_version_records(self):
        """
        Yield RASH version information stored in DB. Latest first.

        :rtype: [VersionRecord]

        """
        keys = ['id', 'rash_version', 'schema_version', 'updated']
        sql = """
        SELECT id, rash_version, schema_version, updated
        FROM rash_info
        ORDER BY id DESC
        """
        with self.connection() as connection:
            for row in connection.execute(sql):
                yield VersionRecord(**dict(zip(keys, row)))

    def update_version_records(self):
        """
        Update rash_info table if necessary.
        """
        from .__init__ import __version__ as version
        with self.connection(commit=True) as connection:
            for vrec in self.get_version_records():
                if (vrec.rash_version == version and
                    vrec.schema_version == schema_version):
                    return  # no need to insert the new one!
            connection.execute(
                'INSERT INTO rash_info (rash_version, schema_version) '
                'VALUES (?, ?)',
                [version, schema_version])

    def import_json(self, json_path, **kwds):
        import json
        with open(json_path) as fp:
            try:
                dct = json.load(fp)
            except ValueError:
                warnings.warn(
                    'Ignoring invalid JSON file at: {0}'.format(json_path))
                return
        self.import_dict(dct, **kwds)

    def import_dict(self, dct, check_duplicate=True):
        crec = CommandRecord(**dct)
        if check_duplicate and nonempty(self.select_by_command_record(crec)):
            return
        with self.connection(commit=True) as connection:
            db = connection.cursor()
            ch_id = self._insert_command_history(db, crec)
            self._isnert_command_environment(db, ch_id, crec.environ)
            self._insert_pipe_status(db, ch_id, crec.pipestatus)

    def _insert_command_history(self, db, crec):
        command_id = self._get_maybe_new_command_id(db, crec.command)
        session_id = self._get_maybe_new_session_id(db, crec.session_id)
        directory_id = self._get_maybe_new_directory_id(db, crec.cwd)
        terminal_id = self._get_maybe_new_terminal_id(db, crec.terminal)
        db.execute(
            '''
            INSERT INTO command_history
                (command_id, session_id, directory_id, terminal_id,
                 start_time, stop_time, exit_code)
            VALUES (?, ?, ?, ?, ?, ?, ?)
            ''',
            [command_id, session_id, directory_id, terminal_id,
             convert_ts(crec.start), convert_ts(crec.stop), crec.exit_code])
        return db.lastrowid

    def _isnert_command_environment(self, db, ch_id, environ):
        self._insert_environ(db, 'command_environment_map', 'ch_id', ch_id,
                             environ)

    def _insert_environ(self, db, table, id_name, ch_id, environ):
        if not environ:
            return
        for (name, value) in environ.items():
            if name is None or value is None:
                continue
            ev_id = self._get_maybe_new_id(
                db, 'environment_variable',
                {'variable_name': name, 'variable_value': value})
            db.execute(
                '''
                INSERT INTO {0}
                    ({1}, ev_id)
                VALUES (?, ?)
                '''.format(table, id_name),
                [ch_id, ev_id])

    def _insert_pipe_status(self, db, ch_id, pipe_status):
        if not pipe_status:
            return
        for (i, code) in enumerate(pipe_status):
            db.execute(
                '''
                INSERT INTO pipe_status_map
                    (ch_id, program_position, exit_code)
                VALUES (?, ?, ?)
                ''',
                [ch_id, i, code])

    def _get_maybe_new_command_id(self, db, command):
        if command is None:
            return None
        return self._get_maybe_new_id(
            db, 'command_list', {'command': command})

    def _get_maybe_new_session_id(self, db, session_long_id):
        if session_long_id is None:
            return None
        return self._get_maybe_new_id(
            db, 'session_history', {'session_long_id': session_long_id})

    def _get_maybe_new_directory_id(self, db, directory):
        if directory is None:
            return None
        directory = normalize_directory(directory)
        return self._get_maybe_new_id(
            db, 'directory_list', {'directory': directory})

    def _get_maybe_new_terminal_id(self, db, terminal):
        if terminal is None:
            return None
        return self._get_maybe_new_id(
            db, 'terminal_list', {'terminal': terminal})

    def _get_maybe_new_id(self, db, table, columns):
        kvlist = list(columns.items())
        values = [v for (_, v) in kvlist]
        sql_select = 'SELECT id FROM "{0}" WHERE {1}'.format(
            table,
            ' AND '.join(map('"{0[0]}" = ?'.format, kvlist)),
        )
        for (id_val,) in db.execute(sql_select, values):
            return id_val
        sql_insert = 'INSERT INTO "{0}" ({1}) VALUES ({2})'.format(
            table,
            ', '.join(map('"{0[0]}"'.format, kvlist)),
            ', '.join('?' for _ in kvlist),
        )
        db.execute(sql_insert, values)
        return db.lastrowid

    def select_by_command_record(self, crec):
        """
        Yield records that matches to `crec`.

        All attributes of `crec` except for `environ` are concerned.

        """
        keys = ['command_history_id', 'command', 'session_history_id',
                'cwd', 'terminal',
                'start', 'stop', 'exit_code']
        sql = """
        SELECT
            command_history.id, CL.command, session_id,
            DL.directory, TL.terminal,
            start_time, stop_time, exit_code
        FROM command_history
        LEFT JOIN command_list AS CL ON command_id = CL.id
        LEFT JOIN directory_list AS DL ON directory_id = DL.id
        LEFT JOIN terminal_list AS TL ON terminal_id = TL.id
        WHERE
            (CL.command   = ? OR (CL.command   IS NULL AND ? IS NULL)) AND
            (DL.directory = ? OR (DL.directory IS NULL AND ? IS NULL)) AND
            (TL.terminal  = ? OR (TL.terminal  IS NULL AND ? IS NULL)) AND
            (start_time   = ? OR (start_time   IS NULL AND ? IS NULL)) AND
            (stop_time    = ? OR (stop_time    IS NULL AND ? IS NULL)) AND
            (exit_code    = ? OR (exit_code    IS NULL AND ? IS NULL))
        """
        desired_row = [
            crec.command, normalize_directory(crec.cwd), crec.terminal,
            convert_ts(crec.start), convert_ts(crec.stop), crec.exit_code]
        params = list(itertools.chain(*zip(desired_row, desired_row)))
        return self._select_rows(CommandRecord, keys, sql, params)

    def search_command_record(self, **kwds):
        """
        Search command history.

        :rtype: [CommandRecord]

        """
        (sql, params, keys) = self._compile_sql_search_command_record(**kwds)
        return self._select_rows(CommandRecord, keys, sql, params)

    def _compile_sql_search_command_record(
            cls, limit, unique,
            match_pattern, include_pattern, exclude_pattern,
            cwd, cwd_glob, cwd_under,
            time_after, time_before, duration_longer_than, duration_less_than,
            include_exit_code, exclude_exit_code, reverse, sort_by,
            session_history_id=None,
            **_):  # SOMEDAY: don't ignore unused kwds to search_command_record
        keys = ['command_history_id', 'command', 'session_history_id',
                'cwd', 'terminal',
                'start', 'stop', 'exit_code']
        columns = ['command_history.id', 'CL.command', 'session_id',
                   'DL.directory', 'TL.terminal',
                   'start_time', 'stop_time', 'exit_code']
        max_index = 5
        assert columns[max_index] == 'start_time'
        params = []
        conditions = []

        if cwd_under:
            cwd_glob.extend(os.path.join(os.path.abspath(p), "*")
                            for p in cwd_under)

        def add_or_match(template, name, args):
            conditions.extend(concat_expr(
                'OR', repeat(template.format(name), len(args))))
            params.extend(args)

        conditions.extend(repeat('glob(?, CL.command)', len(match_pattern)))
        params.extend(match_pattern)
        add_or_match('glob(?, {0})', 'CL.command', include_pattern)
        conditions.extend(repeat('NOT glob(?, CL.command)',
                                 len(exclude_pattern)))
        params.extend(exclude_pattern)

        add_or_match('glob(?, {0})', 'DL.directory', cwd_glob)
        add_or_match('{0} = ?', 'DL.directory',
                     [normalize_directory(os.path.abspath(p)) for p in cwd])

        if time_after:
            conditions.append('DATETIME(start_time) >= ?')
            params.append(time_after)

        if time_before:
            conditions.append('DATETIME(start_time) <= ?')
            params.append(time_before)

        command_duration = (
            '(JULIANDAY(stop_time) - JULIANDAY(start_time)) * 60 * 60 * 24')

        if duration_longer_than:
            conditions.append('({0} >= ?)'.format(command_duration))
            params.append(duration_longer_than)

        if duration_less_than:
            conditions.append('({0} <= ?)'.format(command_duration))
            params.append(duration_less_than)

        add_or_match('{0} = ?', 'exit_code', include_exit_code)
        conditions.extend(repeat('exit_code != ?', len(exclude_exit_code)))
        params.extend(exclude_exit_code)

        if session_history_id:
            conditions.append('(session_id = ?)')
            params.append(session_history_id)

        where = ''
        if conditions:
            where = 'WHERE {0} '.format(" AND ".join(conditions))

        group_by = ''
        if unique:
            columns[max_index] = 'MAX({0})'.format(columns[max_index])
            group_by = 'GROUP BY CL.command '
        elif sort_by == 'command_count':
            # When not using "GROUP BY", `COUNT(*)` yields just one
            # row.  As unique is True by default, `unique=False`
            # should means to ignore `sort_by`
            sort_by = None

        order_by = 'start_time'
        if sort_by:
            if sort_by == 'command_count':
                columns.append('COUNT(*) as command_count')
                keys.append('command_count')
            order_by = sort_by

        order_direction = 'ASC' if reverse else 'DESC'

        sql_limit = ''
        if limit and limit >= 0:
            sql_limit = 'LIMIT ?'
            params.append(limit)

        sql = (
            'SELECT {columns} '
            'FROM command_history '
            'LEFT JOIN command_list AS CL ON command_id = CL.id '
            'LEFT JOIN directory_list AS DL ON directory_id = DL.id '
            'LEFT JOIN terminal_list AS TL ON terminal_id = TL.id '
            '{where}{group_by} '
            'ORDER BY {order_by} {order_direction} '
            '{limit}'
        ).format(
            columns=', '.join(columns),
            where=where,
            group_by=group_by,
            order_by=order_by,
            order_direction=order_direction,
            limit=sql_limit,
        )
        return (sql, params, keys)

    def import_init_dict(self, dct, overwrite=True):
        long_id = dct['session_id']
        srec = SessionRecord(**dct)
        with self.connection(commit=True) as connection:
            db = connection.cursor()
            records = list(self.select_session_by_long_id(long_id))
            if records:
                assert len(records) == 1
                oldrec = records[0]
                if oldrec.start is not None and not overwrite:
                    return
                oldrec.start = srec.start
                sh_id = self._update_session_history(db, oldrec)
            else:
                sh_id = self._insert_session_history(db, srec)
            self._update_session_environ(db, sh_id, srec.environ)

    def import_exit_dict(self, dct, overwrite=True):
        long_id = dct['session_id']
        srec = SessionRecord(**dct)
        with self.connection(commit=True) as connection:
            db = connection.cursor()
            records = list(self.select_session_by_long_id(long_id))
            if records:
                assert len(records) == 1
                oldrec = records[0]
                if oldrec.stop is not None and not overwrite:
                    return
                oldrec.stop = srec.stop
                self._update_session_history(db, oldrec)
            else:
                self._insert_session_history(db, srec)

    def _insert_session_history(self, db, srec):
        db.execute(
            '''
            INSERT INTO session_history
                (session_long_id, start_time, stop_time)
            VALUES (?, ?, ?)
            ''',
            [srec.session_id, convert_ts(srec.start), convert_ts(srec.stop)])
        return db.lastrowid

    def _update_session_history(self, db, srec):
        assert srec.session_history_id is not None
        db.execute(
            '''
            UPDATE session_history
            SET session_long_id=?, start_time=?, stop_time=?
            WHERE id=?
            ''',
            [srec.session_id, convert_ts(srec.start), convert_ts(srec.stop),
             srec.session_history_id])
        return srec.session_history_id

    def _update_session_environ(self, db, sh_id, environ):
        if not environ:
            return
        db.execute('DELETE FROM session_environment_map WHERE sh_id=?',
                   [sh_id])
        self._insert_session_environ(db, sh_id, environ)

    def _insert_session_environ(self, db, sh_id, environ):
        self._insert_environ(db, 'session_environment_map', 'sh_id', sh_id,
                             environ)

    def select_session_by_long_id(self, long_id):
        keys = ['session_history_id', 'session_id', 'start', 'stop']
        sql = """
        SELECT id, session_long_id, start_time, stop_time
        FROM session_history
        WHERE session_long_id = ?
        """
        params = [long_id]
        with self.connection() as connection:
            for row in connection.execute(sql, params):
                yield SessionRecord(**dict(zip(keys, row)))

    def search_session_record(self, session_id):
        return self.select_session_by_long_id(session_id)

    def get_full_command_record(self, command_history_id,
                                merge_session_environ=True):
        """
        Get fully retrieved :class:`CommandRecord` instance by ID.

        By "fully", it means that complex slots such as `environ` and
        `pipestatus` are available.

        :type    command_history_id: int
        :type merge_session_environ: bool

        """
        with self.connection() as db:
            crec = self._select_command_record(db, command_history_id)
            crec.pipestatus = self._get_pipestatus(db, command_history_id)
            # Set environment variables
            cenv = self._select_environ(db, 'command', command_history_id)
            crec.environ.update(cenv)
            if merge_session_environ:
                senv = self._select_environ(
                    db, 'session', crec.session_history_id)
                crec.environ.update(senv)
        return crec

    def _select_command_record(self, db, command_history_id):
        keys = ['session_history_id', 'command', 'cwd', 'terminal',
                'start', 'stop', 'exit_code']
        sql = """
        SELECT
            session_id, CL.command, DL.directory, TL.terminal,
            start_time, stop_time, exit_code
        FROM command_history
        LEFT JOIN command_list AS CL ON command_id = CL.id
        LEFT JOIN directory_list AS DL ON directory_id = DL.id
        LEFT JOIN terminal_list AS TL ON terminal_id = TL.id
        WHERE command_history.id = ?
        """
        params = [command_history_id]
        for row in db.execute(sql, params):
            crec = CommandRecord(**dict(zip(keys, row)))
            crec.command_history_id = command_history_id
            return crec
        raise ValueError("Command record of id={0} is not found"
                         .format(command_history_id))

    def _get_pipestatus(self, db, command_history_id):
        sql = """
        SELECT program_position, exit_code
        FROM pipe_status_map
        WHERE ch_id = ?
        """
        params = [command_history_id]
        records = list(db.execute(sql, params))
        length = max(r[0] for r in records) + 1
        pipestatus = [None] * length
        for (i, s) in records:
            pipestatus[i] = s
        return pipestatus

    def _select_environ(self, db, recname, recid):
        sql = """
        SELECT
            EVar.variable_name, EVar.variable_value
        FROM {recname}_environment_map as EMap
        LEFT JOIN environment_variable AS EVar ON EMap.ev_id = EVar.id
        WHERE EMap.{hist_id} = ?
        """.format(
            recname=recname,
            hist_id='ch_id' if recname == 'command' else 'sh_id',
        )
        params = [recid]
        return db.execute(sql, params)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import subprocess
import unittest
import tempfile
import shutil
import textwrap
import json
import time

from ..utils.py3compat import PY3
from ..config import ConfigStore
from ..tests.utils import BaseTestCase, skipIf

BASE_COMMAND = 'rash'

try:
    run_command = subprocess.check_output
except AttributeError:

    def run_command(*args, **kwds):
        assert 'stdout' not in kwds
        with open(os.devnull, 'w') as devnull:
            kwds['stdout'] = devnull
            subprocess.check_call(*args, **kwds)


def run_cli(command, *args, **kwds):
    run_command([BASE_COMMAND] + command, *args, **kwds)


class TestCLI(unittest.TestCase):

    def test_command_init_known_shell(self):
        run_cli(['init', '--shell', 'zsh'])

    def test_command_init_unknown_shell(self):
        self.assertRaises(
            subprocess.CalledProcessError,
            run_cli,
            ['init', '--shell', 'UNKNOWN_SHELL'], stderr=subprocess.PIPE)


class FunctionalTestMixIn(object):

    """
    MixIn class for isolating functional test environment.

    SOMEDAY: Make FunctionalTestMixIn work in non-*nix systems.
    (I think) This isolation does not work in Mac OS in Windows!
    I can workaround this by adding --config-dir global
    option to specify configuration directory from
    command line, rather than using $HOME.

    """

    def setUp(self):
        self.home_dir = tempfile.mkdtemp(prefix='rash-test-')
        self.config_dir = os.path.join(self.home_dir, '.config')
        self.conf_base_path = os.path.join(self.config_dir, 'rash')

        self.environ = os.environ.copy()
        self.environ['HOME'] = self.home_dir
        # Make sure that $XDG_CONFIG_HOME does not confuse sub processes
        if 'XDG_CONFIG_HOME' in self.environ:
            del self.environ['XDG_CONFIG_HOME']

        self.conf = ConfigStore(self.conf_base_path)

    def tearDown(self):
        # Kill daemon if exists
        try:
            if os.path.exists(self.conf.daemon_pid_path):
                with open(self.conf.daemon_pid_path) as f:
                    pid = f.read().strip()
                print("Daemon (PID={0}) may be left alive.  Killing it..."
                      .format(pid))
                subprocess.call(['kill', pid])
        except Exception as e:
            print("Got error while trying to kill daemon: {0}"
                  .format(e))

        try:
            shutil.rmtree(self.home_dir)
        except OSError:
            print("Failed to remove self.home_dir={0}. "
                  "Can be timing issue.  Trying again..."
                  .format(self.home_dir))
            time.sleep(0.1)
            shutil.rmtree(self.home_dir)

    def popen(self, *args, **kwds):
        if 'env' in kwds:
            raise RuntimeError('Do not use env!')
        if 'cwd' in kwds:
            raise RuntimeError('Do not use cwd!')
        kwds['env'] = self.environ
        kwds['cwd'] = self.home_dir
        return subprocess.Popen(*args, **kwds)


class TestIsolation(FunctionalTestMixIn, BaseTestCase):

    """
    Make sure that test environment is isolated from the real one.
    """

    def test_config_isolation(self):
        proc = self.popen(
            [os.path.abspath(sys.executable)],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate(textwrap.dedent("""
        from rash.config import ConfigStore
        conf = ConfigStore()
        print(repr(conf.base_path))
        """).encode())
        stderr = stderr.decode()
        stdout = stdout.decode()
        base_path = eval(stdout)
        self.assertEqual(base_path, self.conf_base_path)
        self.assertFalse(stderr)
        self.assertNotEqual(base_path, ConfigStore().base_path)


class ShellTestMixIn(FunctionalTestMixIn):

    shell = 'sh'
    eval_command = 'eval'

    def run_shell(self, script):
        proc = self.popen(
            [self.shell],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate(script.encode())
        return (stdout.decode(), stderr.decode())

    def get_record_data(self, record_type):
        top = os.path.join(self.conf.record_path, record_type)
        for (root, _, files) in os.walk(top):
            for f in files:
                path = os.path.join(root, f)
                with open(path) as f:
                    data = json.load(f)
                yield dict(path=path, data=data)

    def get_all_record_data(self):
        return dict(
            init=list(self.get_record_data('init')),
            exit=list(self.get_record_data('exit')),
            command=list(self.get_record_data('command')),
        )

    def _get_init_script(self, no_daemon=True, daemon_options=[],
                        daemon_outfile=None):
        options = []
        if no_daemon:
            options.append('--no-daemon')
        options.extend(map('--daemon-opt={0}'.format, daemon_options))
        if daemon_outfile:
            options.extend(['--daemon-outfile', daemon_outfile])
        optstr = ' '.join(options)
        return "{0} $({1} init --shell {2} {3})".format(
            self.eval_command, BASE_COMMAND, self.shell, optstr)

    def get_script(self, script='', **kwds):
        init_script = self._get_init_script(**kwds)
        return '\n'.join([init_script, textwrap.dedent(script)])

    def test_init(self):
        script = self.get_script("""
        test -n "$_RASH_SESSION_ID" && echo "_RASH_SESSION_ID is defined"
        """)
        (stdout, stderr) = self.run_shell(script)
        self.assertFalse(stderr)
        self.assertIn('_RASH_SESSION_ID is defined', stdout)

        assert os.path.isdir(self.conf.record_path)
        records = self.get_all_record_data()
        self.assertEqual(len(records['init']), 1)
        self.assertEqual(len(records['exit']), 1)
        self.assertEqual(len(records['command']), 0)

        from ..record import get_environ
        subenv = get_environ(['HOST'])

        data = records['init'][0]['data']
        assert 'start' in data
        assert 'stop' not in data
        self.assertEqual(data['environ']['HOST'], subenv['HOST'])
        init_id = data['session_id']

        data = records['exit'][0]['data']
        assert 'start' not in data
        assert 'stop' in data
        assert not data['environ']
        exit_id = data['session_id']

        self.assertEqual(init_id, exit_id)

    def test_postexec(self):
        script = self.get_script(self.test_postexec_script)
        (stdout, stderr) = self.run_shell(script)

        # stderr may have some errors in it
        if stderr:
            print("Got STDERR from {0} (but it's OK to ignore it)"
                  .format(self.shell))
            print(stderr)

        records = self.get_all_record_data()
        self.assertEqual(len(records['init']), 1)
        self.assertEqual(len(records['exit']), 1)
        self.assertEqual(len(records['command']), 1)

        init_data = records['init'][0]['data']
        command_data = records['command'][0]['data']
        assert command_data['session_id'] == init_data['session_id']
        assert command_data['environ']['PATH']
        assert isinstance(command_data['stop'], int)
        if self.shell.endswith('zsh'):
            assert isinstance(command_data['start'], int)
        else:
            assert 'start' not in command_data

    test_postexec_script = None
    """Set this to a shell script for :meth:`test_postexc`."""

    def test_exit_code(self):
        script = self.get_script(self.test_exit_code_script)
        (stdout, stderr) = self.run_shell(script)

        # stderr may have some errors in it
        if stderr:
            print("Got STDERR from {0} (but it's OK to ignore it)"
                  .format(self.shell))
            print(stderr)

        records = self.get_all_record_data()
        self.assertEqual(len(records['init']), 1)
        self.assertEqual(len(records['exit']), 1)
        self.assertEqual(len(records['command']), 1)

        command_data = [d['data'] for d in records['command']]
        self.assertEqual(command_data[0]['exit_code'], 1)

    test_exit_code_script = None
    """Set this to a shell script for :meth:`test_exit_code`."""

    def test_pipe_status(self):
        script = self.get_script(self.test_pipe_status_script)
        (stdout, stderr) = self.run_shell(script)

        # stderr may have some errors in it
        if stderr:
            print("Got STDERR from {0} (but it's OK to ignore it)"
                  .format(self.shell))
            print(stderr)

        records = self.get_all_record_data()
        self.assertEqual(len(records['init']), 1)
        self.assertEqual(len(records['exit']), 1)
        self.assertEqual(len(records['command']), 1)

        command_data = [d['data'] for d in records['command']]
        self.assertEqual(command_data[0]['pipestatus'], [1, 0])

    test_pipe_status_script = None
    """Set this to a shell script for :meth:`test_pipe_status`."""

    def test_non_existing_directory(self):
        main_script = """
        _rash-precmd
        mkdir non_existing_directory

        _rash-precmd
        cd non_existing_directory

        _rash-precmd
        rmdir ../non_existing_directory

        _rash-precmd
        :

        _rash-precmd
        cd ..
        """
        script = self.get_script(main_script)
        (stdout, stderr) = self.run_shell(script)
        self.assertNotIn('Traceback', stderr)

    @skipIf(PY3, "watchdog does not support Python 3")
    def test_daemon(self):
        daemon_outfile = os.path.join(self.conf.base_path, 'daemon.out')
        script = self.get_script(
            no_daemon=False, daemon_outfile=daemon_outfile,
            daemon_options=['--keep-json', '--log-level=DEBUG'])
        (stdout, stderr) = self.run_shell(script)

        # These are useful when debugging, so let's leave them:
        print(stderr)
        print(stdout)
        print(self.conf.daemon_pid_path)

        # Print daemon process output for debugging
        with open(daemon_outfile) as f:
            daemon_output = f.read().strip()
            if daemon_output:
                print("Daemon process output ({0})".format(daemon_outfile))
                print(daemon_output.decode())

        # The daemon process should create a PID file containing a number
        @self.assert_poll_do(
            "Daemon did not produce PID file at: {0}"
            .format(self.conf.daemon_pid_path))
        def pid_file_contains_a_number():
            try:
                with open(self.conf.daemon_pid_path) as f:
                    return f.read().strip().isdigit()
            except IOError:
                return False

        # Read the PID file
        with open(self.conf.daemon_pid_path) as f:
            pid = int(f.read().strip())

        # The daemon process should be alive
        ps_pid_cmd = ['ps', '--pid', str(pid)]
        try:
            run_command(ps_pid_cmd)
        except subprocess.CalledProcessError:
            raise AssertionError(
                'At this point, daemon process should be live '
                '("ps --pid {0}" failed).'.format(pid))

        # The daemon should create a log file
        self.assert_poll(lambda: os.path.exists(self.conf.daemon_log_path),
                         "daemon_log_path={0!r} is not created on time"
                         .format(self.conf.daemon_log_path))

        # The daemon should write some debug message to the log file
        # (Note: --log-level=DEBUG is given by $RASH_INIT_DAEMON_OPTIONS)
        with open(self.conf.daemon_log_path) as f:
            @self.assert_poll_do("Nothing written in log file.")
            def log_file_written():
                return f.read().strip()

        # Kill command should succeeds
        run_command(['kill', '-TERM', str(pid)])

        # The daemon should be killed by the TERM signal
        @self.assert_poll_do(
            "Daemon process {0} failed to exit.".format(pid))
        def terminated():
            try:
                run_command(ps_pid_cmd)
                return False
            except subprocess.CalledProcessError:
                return True

        # The daemon should remove the PID file on exit
        self.assert_poll(
            lambda: not os.path.exists(self.conf.daemon_pid_path),
            "Daemon did not remove PID file at: {0}".format(
                self.conf.daemon_pid_path))

    @staticmethod
    def assert_poll(assertion, message, num=100, tick=0.1):
        """
        Run `assersion` every `tick` second `num` times.

        If none of `assersion` call returns true, it raise
        an assertion error with `message`.

        """
        for i in range(num):
            if assertion():
                break
            time.sleep(tick)
        else:
            raise AssertionError(message)

    @classmethod
    def assert_poll_do(cls, message, *args, **kwds):
        """
        Decorator to run :meth:`assert_poll` right after the definition.
        """
        def decorator(assertion):
            cls.assert_poll(assertion, message, *args, **kwds)
            return assertion
        return decorator


class TestZsh(ShellTestMixIn, BaseTestCase):
    shell = 'zsh'
    test_postexec_script = """
    _rash-precmd
    """
    test_exit_code_script = """
    false
    _rash-precmd
    """
    test_pipe_status_script = """
    false | true
    _rash-precmd
    """

    def test_zsh_executes_preexec(self):
        script = self.get_script('echo _RASH_EXECUTING=$_RASH_EXECUTING')
        (stdout, stderr) = self.run_shell(script)
        self.assertFalse(stderr)
        self.assertIn('_RASH_EXECUTING=t', stdout)

    def test_hook_installation(self):
        script = self.get_script("""
        echo $precmd_functions
        echo $preexec_functions
        """)
        (stdout, stderr) = self.run_shell(script)
        self.assertIn('_rash-precmd', stdout)
        self.assertIn('_rash-preexec', stdout)


class TestBash(ShellTestMixIn, BaseTestCase):
    shell = 'bash'
    test_postexec_script = """
    eval "$PROMPT_COMMAND"
    eval "$PROMPT_COMMAND"
    """
    test_exit_code_script = """
    eval "$PROMPT_COMMAND"
    false
    eval "$PROMPT_COMMAND"
    """
    test_pipe_status_script = """
    eval "$PROMPT_COMMAND"
    false | true
    eval "$PROMPT_COMMAND"
    """

    def test_hook_installation(self):
        script = self.get_script('echo $PROMPT_COMMAND')
        (stdout, stderr) = self.run_shell(script)
        self.assertIn('_rash-precmd', stdout)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import datetime
import itertools

from ..model import CommandRecord, SessionRecord
from ..database import DataBase, normalize_directory
from ..utils.py3compat import nested
from .utils import BaseTestCase, monkeypatch


def setdefaults(d, **kwds):
    for (k, v) in kwds.items():
        d.setdefault(k, v)


def to_sql_timestamp(ts):
    if ts is not None:
        dt = datetime.datetime.utcfromtimestamp(ts)
        return dt.strftime("%Y-%m-%d %H:%M:%S")


def to_record(recclass, data):
    crec = recclass(**data)
    crec.start = to_sql_timestamp(crec.start)
    crec.stop = to_sql_timestamp(crec.stop)
    return crec


def to_command_record(data):
    crec = to_record(CommandRecord, data)
    crec.cwd = normalize_directory(crec.cwd)
    return crec


def to_session_record(data):
    return to_record(SessionRecord, data)


class InMemoryDataBase(DataBase):

    def __init__(self):
        import sqlite3
        db = sqlite3.connect(':memory:')
        self._get_db = lambda: db
        self._init_db()
        self.update_version_records()


class TestInMemoryDataBase(BaseTestCase):

    dbclass = InMemoryDataBase

    def abspath(self, *ps):
        return os.path.join(os.path.sep, *ps)

    def setUp(self):
        self.db = self.dbclass()

    def test_info_get_version(self):
        from ..__init__ import __version__
        from ..database import schema_version
        verrec = next(self.db.get_version_records())
        self.assertEqual(verrec.rash_version, __version__)
        self.assertEqual(verrec.schema_version, schema_version)

    def test_info_version_update(self):
        from .. import __init__
        from .. import database
        new_project_ver = '100.0.0'
        new_schema_ver = '100.0.0'
        with nested(monkeypatch(__init__, '__version__', new_project_ver),
                    monkeypatch(database, 'schema_version', new_schema_ver)):
            self.db.update_version_records()
        records = list(self.db.get_version_records())
        self.assertEqual(len(records), 2)
        self.assertEqual(records[0].rash_version, new_project_ver)
        self.assertEqual(records[0].schema_version, new_schema_ver)
        self.assertEqual(records[1].rash_version, __init__.__version__)
        self.assertEqual(records[1].schema_version, database.schema_version)

    def get_default_search_kwds(self):
        import argparse
        from ..search import search_add_arguments, preprocess_kwds
        parser = argparse.ArgumentParser()
        search_add_arguments(parser)
        kwds = vars(parser.parse_args([]))
        return preprocess_kwds(kwds)

    def search_command_record(self, **kwds):
        setdefaults(kwds, **self.get_default_search_kwds())
        return list(self.db.search_command_record(**kwds))

    def assert_same_command_record(
            self, crec1, crec2,
            keys=['command', 'cwd', 'terminal', 'start', 'stop', 'exit_code']):
        asdict = lambda rec: dict((k, getattr(rec, k)) for k in keys)
        self.assertEqual(asdict(crec1), asdict(crec2))

    def assert_not_same_command_record(self, *args, **kwds):
        self.assertRaises(AssertionError, self.assert_same_command_record,
                          *args, **kwds)

    def get_dummy_command_record_data(self):
        return {
            'command': 'DUMMY COMMAND',
            'cwd': self.abspath('DUMMY', 'WORKING', 'DIRECTORY'),
            'exit_code': 0,
            'pipestatus': [2, 3, 0],
            'start': 100,
            'stop': 102,
            'terminal': 'DUMMY_TERMINAL',
            'session_id': 'DUMMY-SESSION-ID',
            'environ': {
                'PATH': 'DUMMY:PATH:DATA',
            },
        }

    def test_empty_database(self):
        records = self.search_command_record()
        self.assertEqual(len(records), 0)

    def test_import_empty_command_record(self):
        self.db.import_dict({})
        records = self.search_command_record()
        self.assert_same_command_record(records[0], to_command_record({}))
        self.assertEqual(len(records), 1)

    def test_import_command_record(self):
        data = self.get_dummy_command_record_data()
        self.db.import_dict(data)
        records = self.search_command_record()
        crec = records[0]
        self.assert_same_command_record(crec, to_command_record(data))
        self.assertEqual(len(records), 1)

    def test_import_command_record_no_check_duplicate(self):
        data = self.get_dummy_command_record_data()
        num = 3
        for _ in range(num):
            self.db.import_dict(data, check_duplicate=False)
        records = self.search_command_record(unique=False)
        for crec in records:
            self.assert_same_command_record(crec, to_command_record(data))
        self.assertEqual(len(records), num)

    def test_import_command_record_check_duplicate(self):
        data = self.get_dummy_command_record_data()
        # Import different version of data which DB cannot distinguish.
        for env in [{'SHELL': 'bash'}, {'SHELL': 'zsh'},
                    {'SHELL': 'tcsh'}, {'SHELL': 'sh'}]:
            data['environ'].update(env)
            self.db.import_dict(data, check_duplicate=True)
        records = self.search_command_record(unique=False)
        self.assert_same_command_record(records[0], to_command_record(data))
        self.assertEqual(len(records), 1)

    def test_search_command_by_pattern(self):
        data1 = self.get_dummy_command_record_data()
        data2 = self.get_dummy_command_record_data()
        data1['command'] = 'git status'
        data2['command'] = 'hg status'
        self.db.import_dict(data1)
        self.db.import_dict(data2)
        dcrec1 = to_command_record(data1)
        dcrec2 = to_command_record(data2)

        records = self.search_command_record(include_pattern=['git*'],
                                             unique=False)
        crec = records[0]
        self.assert_same_command_record(crec, dcrec1)
        self.assert_not_same_command_record(crec, dcrec2)
        self.assertEqual(len(records), 1)

        records = self.search_command_record(include_pattern=['bzr*'],
                                             unique=False)
        self.assertEqual(len(records), 0)

    def test_search_command_by_exclude_pattern(self):
        data1 = self.get_dummy_command_record_data()
        data2 = self.get_dummy_command_record_data()
        data1['command'] = 'git status'
        data2['command'] = 'hg status'
        self.db.import_dict(data1)
        self.db.import_dict(data2)
        dcrec1 = to_command_record(data1)
        dcrec2 = to_command_record(data2)

        records = self.search_command_record(exclude_pattern=['hg*'],
                                             unique=False)
        crec = records[0]
        self.assert_same_command_record(crec, dcrec1)
        self.assert_not_same_command_record(crec, dcrec2)
        self.assertEqual(len(records), 1)

        records = self.search_command_record(exclude_pattern=['bzr*'],
                                             unique=False)
        self.assertEqual(len(records), 2)

    def test_search_command_by_complex_pattern(self):
        data1 = self.get_dummy_command_record_data()
        data2 = self.get_dummy_command_record_data()
        data1['command'] = 'git status'
        data2['command'] = 'hg status'
        self.db.import_dict(data1)
        self.db.import_dict(data2)
        dcrec1 = to_command_record(data1)

        records = self.search_command_record(
            include_pattern=['*status'], exclude_pattern=['hg*'],
            unique=False)
        self.assertEqual(len(records), 1)
        self.assert_same_command_record(records[0], dcrec1)

        records = self.search_command_record(
            include_pattern=['hg*'], exclude_pattern=['*status'],
            unique=False)
        self.assertEqual(len(records), 0)

    def test_search_command_by_cwd(self):
        data1 = self.get_dummy_command_record_data()
        data2 = self.get_dummy_command_record_data()
        data1['cwd'] = self.abspath('DUMMY', 'A')
        data2['cwd'] = self.abspath('DUMMY', 'B')
        self.db.import_dict(data1)
        self.db.import_dict(data2)
        dcrec1 = to_command_record(data1)
        dcrec2 = to_command_record(data2)

        records = self.search_command_record(cwd=[data1['cwd']], unique=False)
        crec = records[0]
        self.assert_same_command_record(crec, dcrec1)
        self.assert_not_same_command_record(crec, dcrec2)
        self.assertEqual(len(records), 1)

        records = self.search_command_record(cwd=[self.abspath('DUMMY')],
                                             unique=False)
        self.assertEqual(len(records), 0)

    def test_search_command_by_cwd_glob(self):
        data1 = self.get_dummy_command_record_data()
        data2 = self.get_dummy_command_record_data()
        data1['cwd'] = self.abspath('DUMMY', 'A')
        data2['cwd'] = self.abspath('DUMMY', 'B')
        self.db.import_dict(data1)
        self.db.import_dict(data2)
        dcrec1 = to_command_record(data1)
        dcrec2 = to_command_record(data2)

        records = self.search_command_record(
            cwd_glob=[self.abspath('DUMMY', '*')], unique=False)
        self.assert_same_command_record(records[0], dcrec1)
        self.assert_same_command_record(records[1], dcrec2)
        self.assertEqual(len(records), 2)

        records = self.search_command_record(
            cwd_glob=[self.abspath('REAL', '*')], unique=False)
        self.assertEqual(len(records), 0)

    def test_search_command_sort_by_command_count(self):
        command_num_pairs = [('command A', 10),
                             ('command B', 5),
                             ('command C', 15)]
        for (command, num) in command_num_pairs:
            for i in range(num):
                data = self.get_dummy_command_record_data()
                data.update(
                    command=command,
                    # Use `start` to make the record unique
                    start=i)
                self.db.import_dict(data)

        records = self.search_command_record(sort_by='command_count')
        self.assertEqual(len(records), 3)

        commands = [r.command for r in records]
        self.assertEqual(commands, ['command C', 'command A', 'command B'])

        counts = [r.command_count for r in records]
        self.assertEqual(counts, [15, 10, 5])

    def test_search_command_with_connection(self):
        num = 5
        small_num = 3
        for i in range(num):
            data = self.get_dummy_command_record_data()
            data.update(command='DUMMY-COMMAND-{0}'.format(i))
            self.db.import_dict(data)

        kwds = dict(unique=False)
        setdefaults(kwds, **self.get_default_search_kwds())

        # Generator hold connection to DB
        records = self.db.search_command_record(**kwds)
        self.assertEqual(len(list(records)), num)

        # Generator can be resumed
        records = self.db.search_command_record(**kwds)
        first_half = list(itertools.islice(records, small_num))
        second_half = list(records)
        self.assertEqual(len(first_half), small_num)
        self.assertEqual(len(second_half), num - small_num)

        # Generator stops if the connection is closed
        records = self.db.search_command_record(**kwds)
        first_half = list(itertools.islice(records, small_num))
        self.db.close_connection()
        second_half = list(records)
        self.assertEqual(len(first_half), small_num)
        self.assertEqual(len(second_half), 0)

    def search_session_record(self, **kwds):
        return list(self.db.search_session_record(**kwds))

    def assert_same_session_record(
            self, srec1, srec2, keys=['session_id', 'start', 'stop']):
        asdict = lambda rec: dict((k, getattr(rec, k)) for k in keys)
        self.assertEqual(asdict(srec1), asdict(srec2))

    def assert_not_same_session_record(self, *args, **kwds):
        self.assertRaises(AssertionError, self.assert_same_session_record,
                          *args, **kwds)

    def test_import_init_record_and_then_exit_record(self):
        session_id = 'DUMMY-SESSION-ID'
        init_data = {'session_id': session_id, 'start': 100}
        exit_data = {'session_id': session_id, 'stop': 102}
        self.db.import_init_dict(init_data)
        self.db.import_exit_dict(exit_data)

        session_data = {}
        session_data.update(init_data)
        session_data.update(exit_data)

        records = self.search_session_record(session_id=session_id)
        self.assert_same_session_record(records[0],
                                        to_session_record(session_data))
        self.assertEqual(len(records), 1)

    def test_import_exit_record_and_then_init_record(self):
        session_id = 'DUMMY-SESSION-ID'
        init_data = {'session_id': session_id, 'start': 100}
        exit_data = {'session_id': session_id, 'stop': 102}
        self.db.import_exit_dict(exit_data)
        self.db.import_init_dict(init_data)

        session_data = {}
        session_data.update(init_data)
        session_data.update(exit_data)

        records = self.search_session_record(session_id=session_id)
        self.assert_same_session_record(records[0],
                                        to_session_record(session_data))
        self.assertEqual(len(records), 1)

    def test_import_session_record_with_environ(self):
        session_id = 'DUMMY-SESSION-ID'
        init_data = {'session_id': session_id, 'start': 100}
        exit_data = {'session_id': session_id, 'stop': 102}
        init_data['environ'] = {'SHELL': 'zsh'}
        self.db.import_init_dict(init_data)
        self.db.import_exit_dict(exit_data)

        session_data = {}
        session_data.update(init_data)
        session_data.update(exit_data)

        records = self.search_session_record(session_id=session_id)
        self.assert_same_session_record(records[0],
                                        to_session_record(session_data))
        self.assertEqual(len(records), 1)

    def check_consistency_one_command_in_one_session(self):
        session_id = 'DUMMY-SESSION-ID'
        session_records = self.search_session_record(session_id=session_id)
        sh_id = session_records[0].session_history_id
        self.assertIsInstance(sh_id, int)
        command_records = self.search_command_record(session_history_id=sh_id)
        self.assertEqual(sh_id, command_records[0].session_history_id)

    def test_import_session_record_after_command_record(self):
        self.test_import_command_record()
        self.test_import_exit_record_and_then_init_record()
        self.check_consistency_one_command_in_one_session()

    def test_import_session_record_before_command_record(self):
        self.test_import_exit_record_and_then_init_record()
        self.test_import_command_record()
        self.check_consistency_one_command_in_one_session()

    def test_get_full_command_record_simple_keys(self):
        command_data = self.get_dummy_command_record_data()
        self.db.import_dict(command_data)

        records = self.search_command_record()
        self.assertEqual(len(records), 1)
        command_history_id = records[0].command_history_id

        crec = self.db.get_full_command_record(command_history_id)
        self.assert_same_command_record(crec, to_command_record(command_data))

    def test_get_full_command_record_merging_session_environ(self):
        session_id = 'DUMMY-SESSION-ID'
        init_data = {'session_id': session_id, 'start': 100}
        init_data['environ'] = {'SHELL': 'zsh'}
        command_data = self.get_dummy_command_record_data()
        command_data['session_id'] = session_id

        desired_environ = {}
        desired_environ.update(command_data['environ'])
        desired_environ.update(init_data['environ'])

        self.db.import_dict(command_data)
        self.db.import_init_dict(init_data)

        records = self.search_command_record()
        self.assertEqual(len(records), 1)
        command_history_id = records[0].command_history_id

        crec = self.db.get_full_command_record(command_history_id)
        self.assertEqual(crec.environ, desired_environ)

        crec = self.db.get_full_command_record(command_history_id,
                                               merge_session_environ=False)
        self.assertEqual(crec.environ, command_data['environ'])

    def test_get_full_command_record_no_environ_leak(self):
        session_id_1 = 'DUMMY-SESSION-ID-1'
        session_id_2 = 'DUMMY-SESSION-ID-2'
        init_data_1 = {'session_id': session_id_1, 'start': 100}
        init_data_2 = {'session_id': session_id_2, 'start': 100}
        init_data_1['environ'] = {'SHELL': 'zsh'}
        init_data_2['environ'] = {'SHELL': 'bash'}
        command_data_1 = self.get_dummy_command_record_data()
        command_data_2 = self.get_dummy_command_record_data()
        command_data_1['session_id'] = session_id_1
        command_data_2['session_id'] = session_id_2

        desired_environ = {}
        desired_environ.update(command_data_1['environ'])
        desired_environ.update(init_data_1['environ'])

        self.db.import_dict(command_data_1)
        self.db.import_dict(command_data_2)
        self.db.import_init_dict(init_data_1)
        self.db.import_init_dict(init_data_2)

        records = list(self.db.select_by_command_record(
            to_command_record(command_data_1)))
        self.assertEqual(len(records), 1)
        command_history_id = records[0].command_history_id

        crec = self.db.get_full_command_record(command_history_id)
        self.assertEqual(crec.environ, desired_environ)

    def test_get_full_command_record_pipestatus(self):
        command_data = self.get_dummy_command_record_data()
        command_data['pipestatus'] = [2, 3, 0]
        self.db.import_dict(command_data)

        records = self.search_command_record()
        self.assertEqual(len(records), 1)
        command_history_id = records[0].command_history_id

        crec = self.db.get_full_command_record(command_history_id)
        self.assertEqual(crec.pipestatus, command_data['pipestatus'])



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class CommandRecord(object):

    """
    Command record.

    >>> CommandRecord()
    <CommandRecord: None(?.?)>
    >>> CommandRecord(
    ... command='DUMMY-COMMAND',
    ... command_history_id=222,
    ... session_history_id=111)
    <CommandRecord: DUMMY-COMMAND(111.222)>

    """

    def __init__(self, **kwds):
        self.command_history_id = None
        self.command = None
        self.session_history_id = None
        self.session_id = None
        self.exit_code = None
        self.pipestatus = []
        self.start = None
        self.stop = None
        self.terminal = None
        self.cwd = None
        self.environ = {}

        # Non-record attributes (metadata):
        self.command_count = None

        self.__dict__.update(kwds)

    def __repr__(self):
        ch_id = self.command_history_id
        sh_id = self.session_history_id
        return '<{0}: {1}({2}.{3})>'.format(
            self.__class__.__name__,
            self.command,
            sh_id if sh_id is not None else '?',
            ch_id if ch_id is not None else '?',
        )


class SessionRecord(object):

    """
    Session record.

    >>> SessionRecord()
    <SessionRecord: ?>
    >>> SessionRecord(command='DUMMY-COMMAND', session_history_id=111)
    <SessionRecord: 111>

    """

    def __init__(self, **kwds):
        self.session_history_id = None
        self.session_id = None
        self.start = None
        self.stop = None
        self.environ = {}
        self.__dict__.update(kwds)

    def __repr__(self):
        sh_id = self.session_history_id
        return '<{0}: {1}>'.format(
            self.__class__.__name__,
            sh_id if sh_id is not None else '?',
        )


class VersionRecord(object):

    """
    Version record.

    >>> VersionRecord(rash_version='0.2.0', schema_version='0.1')
    <VersionRecord: schema=0.1, rash=0.2.0>

    """

    def __init__(self, **kwds):
        self.id = None
        self.rash_version = None
        self.schema_version = None
        self.updated = None
        self.__dict__.update(kwds)

    def __repr__(self):
        return '<{0}: schema={1}, rash={2}>'.format(
            self.__class__.__name__,
            self.schema_version,
            self.rash_version,
        )



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
import datetime

try:
    from parsedatetime import parsedatetime
    HAS_PARSEDATETIME = True
except:
    HAS_PARSEDATETIME = False


def parse_datetime(string):
    if not HAS_PARSEDATETIME:
        return
    cal = parsedatetime.Calendar()
    dates = cal.parse(string)
    if dates:
        return datetime.datetime.utcfromtimestamp(time.mktime(dates[0]))


def parse_duration(string):
    """
    Parse human readable duration.

    >>> parse_duration('1m')
    60
    >>> parse_duration('7 days') == 7 * 24 * 60 * 60
    True

    """
    if string.isdigit():
        return int(string)
    try:
        return float(string)
    except ValueError:
        pass
    string = string.rstrip()
    for (suf, mult) in DURATION_SUFFIX_MAP.items():
        if string.lower().endswith(suf):
            try:
                return parse_duration(string[:-len(suf)].strip()) * mult
            except TypeError:
                return


DURATION_SUFFIX_MAP = {
    'minute': 60,
    'hour': 60 * 60,
    'day': 60 * 60 * 24,
}


def _add_duration():
    dsm = DURATION_SUFFIX_MAP
    additional = {}
    for (suf, mult) in dsm.items():
        additional[suf[0]] = mult
        additional[suf + 's'] = mult
    dsm['min'] = dsm['minute']
    dsm.update(additional)

_add_duration()



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
import shlex

try:
    from percol.finder import FinderMultiQueryString
    assert FinderMultiQueryString  # fool pyflakes
except ImportError:
    # Dummy class for making this module importable:
    FinderMultiQueryString = object

from argparse import ArgumentParser
from .search import search_add_arguments, preprocess_kwds


def strip_glob(string, split_str=' '):
    """
    Strip glob portion in `string`.

    >>> strip_glob('*glob*like')
    'glob like'
    >>> strip_glob('glob?')
    'glo'
    >>> strip_glob('glob[seq]')
    'glob'
    >>> strip_glob('glob[!seq]')
    'glob'

    :type string: str
    :rtype: str

    """
    string = _GLOB_PORTION_RE.sub(split_str, string)
    return string.strip()

_GLOB_PORTION_RE = re.compile(r'\*|.\?|\[[^\]]+\]')


class SafeArgumentParser(ArgumentParser):

    def exit(self, *_, **__):
        raise ValueError

    def print_usage(self, *_):
        pass

    print_help = print_version = print_usage


class RashFinder(FinderMultiQueryString):

    base_query = []

    def __init__(self, *args, **kwds):
        super(RashFinder, self).__init__(*args, **kwds)

        self.__parser = parser = SafeArgumentParser()
        search_add_arguments(parser)

    # Generator should be terminated in order to close connection to
    # sqlite.  Otherwise, sqlite3 modules raise an error saying that
    # it doesn't support multi-threading access.
    lazy_finding = False

    def find(self, query, collection=None):
        try:
            # shlex < 2.7.3 does not work with unicode:
            args = self.base_query + shlex.split(query.encode())
            ns = self.__parser.parse_args(args)
            kwds = preprocess_kwds(vars(ns))
        except (ValueError, SyntaxError):
            return super(RashFinder, self).find(query, collection)

        # SOMEDAY: get rid of this hard-coded search limit by making
        # `search_command_record` thread-safe and setting
        # `lazy_finding = True`.
        kwds['limit'] = 50 if kwds['match_pattern'] else 1000

        # SOMEDAY:  Implement AND in `search_command_record` for `isearch`.
        # `search_command_record` does OR match while percol does AND
        # match.  This means `search_command_record` generate wasted
        # candidates more when there are more patterns.
        records = self.db.search_command_record(**kwds)
        self.collection = collection = (r.command for r in records)

        # There will be no filtering in the super class.
        # I am using it for highlighting matches.
        subquery = self.split_str.join(strip_glob(q, self.split_str)
                                       for q in kwds['match_pattern'])
        return super(RashFinder, self).find(subquery, collection)


def load_rc(percol, path=None, encoding=None):
    import os
    from percol import debug
    if path is None:
        path = os.path.expanduser("~/.percol.d/rc.py")
    try:
        with open(path, 'r') as rc:
            exec(rc.read().decode(encoding or 'utf-8'), locals())
    except Exception as e:
        debug.log("exception", e)


def launch_isearch(conf, rcfile=None, input_encoding=None, base_query=[],
                   **kwds):
    from percol import Percol
    from percol import tty
    import percol.actions as actions

    from .database import DataBase

    # Pass db instance to finder.  Not clean but works and no harm.
    RashFinder.db = DataBase(conf.db_path)
    RashFinder.base_query = base_query

    ttyname = tty.get_ttyname()
    with open(ttyname, "r+w") as tty_f:
        with Percol(descriptors=tty.reconnect_descriptors(tty_f),
                    finder=RashFinder,
                    actions=(actions.output_to_stdout,),
                    # This will be used if the first call for RashFinder.find
                    # fails to fetch collections from DB.
                    candidates=[],
                    **kwds) as percol:
            load_rc(percol, rcfile, input_encoding)
            exit_code = percol.loop()

    exit(exit_code)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


SORT_KEY_SYNONYMS = {
    'count': 'command_count',
    'time': 'start_time',
    'start': 'start_time',
    'stop': 'stop_time',
    'code': 'exit_code',
}


def search_run(output, format, with_command_id, with_session_id, format_level,
               **kwds):
    """
    Search command history.

    """
    from .config import ConfigStore
    from .database import DataBase

    if format_level >= 3:
        format = ("{session_history_id:>5}  "
                  "{command_history_id:>5}  "
                  "{command_count:>4}  {command}\n")
    elif format_level == 2 or with_command_id and with_session_id:
        format = ("{session_history_id:>5}  "
                  "{command_history_id:>5}  {command}\n")
    elif format_level == 1 or with_command_id:
        format = "{command_history_id:>5}  {command}\n"
    elif with_session_id:
        format = "{session_history_id:>5}  {command}\n"
    else:
        format = format.decode('string_escape')

    db = DataBase(ConfigStore().db_path)
    for crec in db.search_command_record(**preprocess_kwds(kwds)):
        output.write(format.format(**crec.__dict__))


def preprocess_kwds(kwds):
    """
    Preprocess keyword arguments for `DataBase.search_command_record`.
    """
    from .utils.timeutils import parse_datetime, parse_duration

    for key in ['output', 'format', 'with_command_id', 'with_session_id']:
        kwds.pop(key, None)

    for key in ['time_after', 'time_before']:
        val = kwds[key]
        if val:
            dt = parse_datetime(val)
            if dt:
                kwds[key] = dt

    for key in ['duration_longer_than', 'duration_less_than']:
        val = kwds[key]
        if val:
            dt = parse_duration(val)
            if dt:
                kwds[key] = dt

    # interpret "pattern" (currently just copying to --include-pattern)
    less_strict_pattern = list(map("*{0}*".format, kwds.pop('pattern', [])))
    kwds['match_pattern'] = kwds['match_pattern'] + less_strict_pattern

    kwds['sort_by'] = SORT_KEY_SYNONYMS[kwds['sort_by']]
    return kwds


def search_add_arguments(parent_parser):
    import argparse
    # Filter
    parser = parent_parser.add_argument_group('Filter')
    parser.add_argument(
        'pattern', nargs='*',
        help="""
        Glob pattern to match substring of command.  It is as same as
        --match-pattern/-m except that the pattern is going to be
        wrapped by `*`s.  If you want to use strict glob pattern
        that matches to entire command, use --match-pattern/-m.
        """)
    parser.add_argument(
        '--match-pattern', '-m', action='append', default=[],
        help="""
        Only commands that match to this glob pattern are listed.
        Unlike --include-pattern/-g, applying this option multiple
        times does AND match.
        """)
    parser.add_argument(
        '--include-pattern', '-g', action='append', default=[],
        help='glob patterns that matches to commands to include.')
    parser.add_argument(
        '--exclude-pattern', '-G', action='append', default=[],
        help='glob patterns that matches to commands to exclude.')
    parser.add_argument(
        '--match-regexp', '-M', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        Only commands that matches to this grep pattern are listed.
        Unlike --include-regexp/-e, applying this option multiple
        times does AND match.
        """)
    parser.add_argument(
        '--include-regexp', '-e', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        Regular expression patterns that matches to commands to include.
        """)
    parser.add_argument(
        '--exclude-regexp', '-E', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        Regular expression patterns that matches to commands to exclude.
        """)
    parser.add_argument(
        '--cwd', '-d', action='append', default=[],
        help="""
        The working directory at the time when the command was run.
        When given several times, items that match to one of the
        directory are included in the result.
        """)
    parser.add_argument(
        '--cwd-glob', '-D', action='append', default=[],
        help="""
        Same as --cwd but it accepts glob expression.
        """)
    parser.add_argument(
        '--cwd-under', '-u', action='append', default=[],
        help="""
        Same as --cwd but include all subdirectories.
        """)
    parser.add_argument(
        '--time-after', '-t',
        help='commands run after the given time')
    parser.add_argument(
        '--time-before', '-T',
        help='commands run before the given time')
    parser.add_argument(
        '--duration-longer-than', '-S',
        help='commands that takes longer than the given time')
    parser.add_argument(
        '--duration-less-than', '-s',
        help='commands that takes less than the given time')
    parser.add_argument(
        '--include-exit-code', '-x', action='append', default=[], type=int,
        help='include command which finished with given exit code.')
    parser.add_argument(
        '--exclude-exit-code', '-X', action='append', default=[], type=int,
        help='exclude command which finished with given exit code.')
    parser.add_argument(
        '--include-session', '-n', action='append', default=[], type=int,
        help="""
        [NOT IMPLEMENTED]
        include command which is issued in given session.
        """)
    parser.add_argument(
        '--exclude-session', '-N', action='append', default=[], type=int,
        help="""
        [NOT IMPLEMENTED]
        exclude command which is issued in given session.
        """)
    parser.add_argument(
        '--include-environ-pattern', '-v', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        include command which associated with environment variable
        that matches to given glob pattern.""")
    parser.add_argument(
        '--exclude-environ-pattern', '-V', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        exclude command which associated with environment variable
        that matches to given glob pattern.""")
    parser.add_argument(
        '--include-environ-regexp', '-w', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        include command which associated with environment variable
        that matches to given glob pattern.""")
    parser.add_argument(
        '--exclude-environ-regexp', '-W', action='append', default=[],
        help="""
        [NOT IMPLEMENTED]
        exclude command which associated with environment variable
        that matches to given glob pattern.""")
    # "global" filters
    parser.add_argument(
        '--limit', '-l', type=int, default=10,
        help='maximum number of history to show. -1 means not limit.')
    parser.add_argument(
        '--no-unique', '-a', dest='unique', action='store_false', default=True,
        help="""
        Include all duplicates.
        """)
    parser.add_argument(
        '--ignore-case', '-i',
        help="""
        [NOT IMPLEMENTED]
        Do case insensitive search.
        """)

    # Sorter
    parser = parent_parser.add_argument_group('Sorter')
    parser.add_argument(
        '--reverse', '-r', action='store_true', default=False,
        help="""
        Reverse order of the result.
        By default, most recent commands are shown.
        """)
    parser.add_argument(
        '--sort-by', default='count',
        choices=sorted(SORT_KEY_SYNONYMS),
        help="""
        Sort keys
        `count`: number of the time command is executed;
        `start`(=`time`): the time command is executed;
        `stop`: the time command is finished;
        `code`: exit code of the command;
        Note that --sort-by=count cannot be used with --no-unique.
        """)
    parser.add_argument(
        '--sort-by-program-frequency',
        help="""
        [NOT IMPLEMENTED]
        most used program comes first.
        """)

    # Modifier
    parser = parent_parser.add_argument_group('Modifier')
    parser.add_argument(
        '--after-context', '-A', type=int, metavar='NUM',
        help="""
        [NOT IMPLEMENTED]
        Print NUM commands executed after matching commands.
        """)
    parser.add_argument(
        '--before-context', '-B', type=int, metavar='NUM',
        help="""
        [NOT IMPLEMENTED]
        Print NUM commands executed before matching commands.
        """)
    parser.add_argument(
        '--context', '-C', type=int, metavar='NUM',
        help="""
        [NOT IMPLEMENTED]
        Print NUM commands executed before and after matching commands.
        """)
    parser.add_argument(
        '--context-type', choices=['session', 'time'],
        help="""
        [NOT IMPLEMENTED]
        `session`: commands executed in the same shell session;
        `time`: commands executed around the same time;
        """)

    # Formatter
    parser = parent_parser.add_argument_group('Formatter')
    parser.add_argument(
        '--with-command-id', action='store_true', default=False,
        help="""
        Print command ID number.
        When this is set, --format option has no effect.
        If --with-session-id is also specified, session ID comes
        at the first column then command ID comes the next column.
        """)
    parser.add_argument(
        '--with-session-id', action='store_true', default=False,
        help="""
        Print session ID number.
        When this is set, --format option has no effect.
        See also: --with-command-id
        """)
    parser.add_argument(
        '--format', default=r'{command}\n',
        help="""
        Python string formatter.  Available keys:
        command, exit_code, pipestatus (a list), start, stop, cwd,
        command_history_id, session_history_id.
        See also:
        http://docs.python.org/library/string.html#format-string-syntax
        """)
    parser.add_argument(
        '-f', dest='format_level', action='count', default=0,
        help="""
        Set formatting detail.  This can be given multiple times to
        make more detailed output.  For example, giving it once
        equivalent to passing --with-command-id and one more -f means
        adding --with-session-id.
        """)

    # Misc
    parser = parent_parser.add_argument_group('Misc')
    parser.add_argument(
        '--output', default='-', type=argparse.FileType('w'),
        help="""
        Output file to write the results in. Default is stdout.
        """)


commands = [
    ('search', search_add_arguments, search_run),
]



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import platform


def get_config_directory(appname):
    """
    Get OS-specific configuration directory.

    :type appname: str
    :arg  appname: capitalized name of the application

    """
    if platform.system().lower() == 'windows':
        path = os.path.join(os.getenv('APPDATA') or '~', appname, appname)
    elif platform.system().lower() == 'darwin':
        path = os.path.join('~', 'Library', 'Application Support', appname)
    else:
        path = os.path.join(os.getenv('XDG_CONFIG_HOME') or '~/.config',
                            appname.lower())
    return os.path.expanduser(path)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os

from .utils.confutils import get_config_directory
from .utils.pathutils import mkdirp


class ConfigStore(object):

    """
    Configuration and data file store.

    RASH stores data in the following directory in Linux::

      * ~/.config/               # $XDG_CONFIG_HOME
      `--* rash/                 # base_path
         |--* daemon.pid         # PID of daemon process
         |--* daemon.log         # Log file for daemon
         `--* data/              # data_path
            |--* db.sqlite       # db_path ("indexed" record)
            `--* record/         # record_path ("raw" record)
               |--* command/     # command log
               `--* init/        # initialization log

    In Mac OS and Windows, :attr:`base_path` may be different but
    structure in the directory is the same.

    """

    def __init__(self, base_path=None):
        self.base_path = base_path or get_config_directory('RASH')
        """
        Root directory for any RASH related data files (``~/.config/rash``).
        """

        self.data_path = os.path.join(self.base_path, 'data')
        """
        Directory to store data collected by RASH (``~/.config/rash/data``).
        """

        self.record_path = os.path.join(self.data_path, 'record')
        """
        Shell history is stored in this directory at the first stage.
        """

        self.db_path = os.path.join(self.data_path, 'db.sqlite')
        """
        Shell history is stored in the DB at this path.
        """

        self.daemon_pid_path = os.path.join(self.base_path, 'daemon.pid')
        """
        A file to store daemon PID (``~/.config/rash/daemon.pid``).
        """

        self.daemon_log_path = os.path.join(self.base_path, 'daemon.log')
        """
        Daemon log file (``~/.config/rash/daemon.log``).
        """

        self.daemon_log_level = 'INFO'
        """
        Daemon log level.
        """

        mkdirp(self.record_path)



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os


def daemon_run(no_error, restart, record_path, keep_json, check_duplicate,
               log_level):
    """
    Run RASH index daemon.

    This daemon watches the directory ``~/.config/rash/data/record``
    and translate the JSON files dumped by ``record`` command into
    sqlite3 DB at ``~/.config/rash/data/db.sqlite``.

    ``rash init`` will start RASH automatically by default.
    But there are alternative ways to start daemon.

    If you want to organize background process in one place such
    as supervisord_, it is good to add `--restart` option to force
    stop other daemon process if you accidentally started it in
    other place.  Here is an example of supervisord_ setup::

      [program:rash-daemon]
      command=rash daemon --restart

    .. _supervisord: http://supervisord.org/

    Alternatively, you can call ``rash index`` in cron job to
    avoid using daemon.  It is useful if you want to use RASH
    on NFS, as it looks like watchdog does not work on NFS.::

      # Refresh RASH DB every 10 minutes
      */10 * * * * rash index

    """
    # Probably it makes sense to use this daemon to provide search
    # API, so that this daemon is going to be the only process that
    # is connected to the DB?
    from .config import ConfigStore
    from .indexer import Indexer
    from .log import setup_daemon_log_file
    from .watchrecord import watch_record, install_sigterm_handler

    install_sigterm_handler()
    conf = ConfigStore()
    if log_level:
        conf.daemon_log_level = log_level

    # SOMEDAY: make PID checking/writing atomic if possible
    if os.path.exists(conf.daemon_pid_path):
        if no_error:
            return
        with open(conf.daemon_pid_path, 'rt') as f:
            pid = int(f.read().strip())
        if restart:
            stop_running_daemon(conf, pid)
        else:
            raise RuntimeError(
                'There is already a running daemon (PID={0})!'.format(pid))

    with open(conf.daemon_pid_path, 'w') as f:
        f.write(str(os.getpid()))

    try:
        setup_daemon_log_file(conf)
        indexer = Indexer(conf, check_duplicate, keep_json, record_path)
        indexer.index_all()
        watch_record(indexer)
    finally:
        os.remove(conf.daemon_pid_path)


def stop_running_daemon(conf, pid):
    import time
    import signal
    os.kill(pid, signal.SIGTERM)
    for _ in range(30):
        time.sleep(0.1)
        if not os.path.exists(conf.daemon_pid_path):
            break
    else:
        raise RuntimeError(
            'Failed to stop running daemon process (PID={0})'
            .format(pid))


def start_daemon_in_subprocess(options, outpath=os.devnull):
    """
    Run `rash daemon --no-error` in background.

    :type options: list of str
    :arg  options: options for "rash daemon" command
    :type outpath: str
    :arg  outpath: path to redirect daemon output

    """
    import subprocess
    import sys
    from .utils.py3compat import nested
    from .utils.pathutils import mkdirp
    if outpath != os.devnull:
        mkdirp(os.path.dirname(outpath))
    with nested(open(os.devnull),
                open(outpath, 'w')) as (stdin, stdout):
        subprocess.Popen(
            [os.path.abspath(sys.executable), '-m', 'rash.cli',
             'daemon', '--no-error'] + options,
            preexec_fn=os.setsid,
            stdin=stdin, stdout=stdout, stderr=subprocess.STDOUT)


def daemon_add_arguments(parser):
    parser.add_argument(
        '--no-error', action='store_true', default=False,
        help="""
        Do nothing if a daemon is already running.
        """)
    parser.add_argument(
        '--restart', action='store_true', default=False,
        help="""
        Kill already running daemon process if exist.
        """)
    parser.add_argument(
        '--record-path',
        help="""
        specify the directory that has JSON records.
        """)
    parser.add_argument(
        '--keep-json', default=False, action='store_true',
        help="""
        Do not remove old JSON files.  It turns on --check-duplicate.
        """)
    parser.add_argument(
        '--check-duplicate', default=False, action='store_true',
        help='do not store already existing history in DB.')
    parser.add_argument(
        '--log-level',
        choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'],
        help='logging level.')


commands = [
    ('daemon', daemon_add_arguments, daemon_run),
]



# Copyright (C) 2013-  Takafumi Arakaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

from ..termdetection import detect_terminal


class TestTerminalDetection(unittest.TestCase):

    def assert_terminal(self, terminal, environ):
        self.assertEqual(detect_terminal(_environ=environ), terminal)

    def test_tmux(self):
        self.assert_terminal(
            'tmux',
            {'TMUX': 'some value', 'TERM': 'screen'})

    def test_byobu(self):
        self.assert_terminal(
            'byobu',
            {'BYOBU_WINDOWS': 'some/path', 'TERM': 'screen'})

    def test_screen(self):
        self.assert_terminal(
            'screen',
            {'TERM': 'screen', 'COLORTERM': 'gnome-terminal'})

    def test_gnome_terminal(self):
        self.assert_terminal(
            'gnome-terminal',
            {'COLORTERM': 'gnome-terminal', 'TERM': 'xterm-color'})

    def test_fallback(self):
        self.assert_terminal('xterm', {'TERM': 'xterm'})



import os
import sys

from distutils.core import setup

import rash

PY3 = (sys.version_info[0] >= 3)

install_requires = [
    'argparse',
]
if not PY3:
    install_requires.extend([
        'parsedatetime',
        'watchdog',
    ])


setup(
    name='rash',
    version=rash.__version__,
    packages=['rash', 'rash.utils',
              'rash.tests', 'rash.utils.tests', 'rash.functional_tests'],
    package_data={
        # See also ./MANIFEST.in
        'rash': [
            'schema.sql',
            os.path.join('ext', '*sh'),
        ],
    },
    author=rash.__author__,
    author_email='aka.tkf@gmail.com',
    url='https://github.com/tkf/rash',
    license=rash.__license__,
    description='Rash Advances Shell History',
    long_description=rash.__doc__,
    keywords='history, shell, search',
    classifiers=[
        "Development Status :: 3 - Alpha",
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Unix Shell",
        "Topic :: System :: Shells",
        "Topic :: System :: Logging"
        # see: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    ],
    install_requires=install_requires,
    entry_points={
        'console_scripts': ['rash = rash.cli:main'],
    },
)
